"""
The «python» feature contains all the base logic associated to managing python package.

Overview
::::::::

.. code-block:: python

    from medikit import require

    # load python feature (idempotent).
    python = require('python')

    # configure our package.
    python.setup(
        name = 'awesome-library',
        author = 'Chuck Norris',
    )

    # add base and extras requirements, with "loose" versionning (the frozen version fits better in requirements*.txt)
    python.add_requirements(
        'django',
        dev=[
            'pytest',
        ],
    )

"""

import itertools
import os
import tempfile
from getpass import getuser

from pip._vendor.distlib.util import parse_requirement
from pip.req import parse_requirements
from piptools.repositories import PyPIRepository
from piptools.resolver import Resolver
from piptools.scripts.compile import get_pip_command
from piptools.utils import format_requirement

from medikit.events import subscribe
from medikit.feature import Feature, ABSOLUTE_PRIORITY


class PythonConfig(Feature.Config):
    """ Configuration API for the «python» feature. """

    def __init__(self):
        self._setup = {}
        self._install_requires = {}
        self._extras_require = {'dev': {}}

    def get(self, item):
        if item == 'install_requires':
            return list(self.get_requirements())
        if item == 'extras_require':
            return {extra: list(self.get_requirements(extra=extra)) for extra in sorted(self._extras_require)}
        return self._setup[item]

    def add_requirements(self, *reqs, **kwargs):
        for req in reqs:
            req = parse_requirement(req)
            if req.name in self._install_requires:
                raise ValueError('duplicate requirement for {}'.format(req.name))
            self._install_requires[req.name] = req

        for extra, reqs in kwargs.items():
            if not extra in self._extras_require:
                self._extras_require[extra] = {}
            for req in reqs:
                req = parse_requirement(req)
                if req.name in self._extras_require[extra]:
                    raise ValueError('duplicate requirement for {}'.format(req.name))
                self._extras_require[extra][req.name] = req

        return self

    def get_extras(self):
        return self._extras_require.keys()

    def get_requirements(self, extra=None):
        reqs = self._install_requires if extra is None else self._extras_require[extra]
        for req in sorted(reqs):
            yield reqs[req].requirement

    def setup(self, **kwargs):
        self._setup.update(kwargs)
        return self

    def get_setup(self):
        return self._setup


class PythonFeature(Feature):
    """
    Adds the requirements/requirements-dev logic, virtualenv support, setup.py generation, a few metadata files used by
    setuptools...

    """

    requires = {'make'}

    Config = PythonConfig

    @subscribe('medikit.feature.make.on_generate', priority=ABSOLUTE_PRIORITY)
    def on_make_generate(self, event):
        """
        **Environment variables**

        - ``PACKAGE``
        - ``PYTHON``
        - ``PYTHON_BASENAME``
        - ``PYTHON_DIR``
        - ``PYTHON_REQUIREMENTS_FILE``
        - ``PYTHON_REQUIREMENTS_DEV_FILE``
        
        **Shortcuts**
        - ``PIP``
        - ``PIP_INSTALL_OPTIONS``

        **Make targets**

        - ``install``
        - ``install-dev``

        :param MakefileEvent event:

        """
        # Python related environment
        event.makefile.updateleft(
            (
                'PACKAGE',
                event.package_name,
            ),
            (
                'PYTHON',
                '$(shell which python)',
            ),
            (
                'PYTHON_BASENAME',
                '$(shell basename $(PYTHON))',
            ),
            (
                'PYTHON_DIRNAME',
                '$(shell dirname $(PYTHON))',
            ),
            (
                'PYTHON_REQUIREMENTS_FILE',
                'requirements.txt',
            ),
            (
                'PYTHON_REQUIREMENTS_DEV_FILE',
                'requirements-dev.txt',
            ),
        )

        event.makefile['PIP'] = '$(PYTHON_DIRNAME)/pip'
        event.makefile['PIP_INSTALL_OPTIONS'] = ''

        event.makefile.get_target('install').install = [
            '$(PIP) install -U pip wheel $(PIP_INSTALL_OPTIONS) -r $(PYTHON_REQUIREMENTS_FILE)'
        ]

        event.makefile.get_target('install-dev').install = [
            '$(PIP) install -U pip wheel $(PIP_INSTALL_OPTIONS) -r $(PYTHON_REQUIREMENTS_DEV_FILE)'
        ]

    @subscribe('medikit.on_start', priority=ABSOLUTE_PRIORITY)
    def on_start(self, event):
        """
        **Events**

        - ``medikit.feature.python.on_generate`` (with the same ``ProjectEvent`` we got, todo: why not deprecate
          it in favor of higher priority medikit.on_start?)

        **Files**

        - ``<yourpackage>/__init__.py``
        - ``MANIFEST.in``
        - ``README.rst``
        - ``classifiers.txt``
        - ``requirements-dev.txt``
        - ``requirements.txt``
        - ``setup.cfg``
        - ``setup.py`` (overwritten)
        - ``version.txt``

        :param ProjectEvent event:
        """
        self.dispatcher.dispatch(__name__ + '.on_generate', event)

        # Some metadata that will prove useful.
        self.render_empty_files('classifiers.txt', 'README.rst')
        self.render_file_inline('MANIFEST.in', 'include *.txt')
        self.render_file_inline(
            'setup.cfg', '''
                    [bdist_wheel]
                    # This flag says that the code is written to work on both Python 2 and Python
                    # 3. If at all possible, it is good practice to do this. If you cannot, you
                    # will need to generate wheels for each Python version that you support.
                    universal=1
                    
                    [metadata]
                    description-file = README.rst
                '''
        )

        # Explode package name so we know which python packages are namespaced and
        # which are not.
        package_bits = event.setup['name'].split('.')
        if len(package_bits) > 1:
            event.setup['namespace_packages'] = []
            for i in range(1, len(package_bits)):
                # TODO convert to string type (six?) depending on python version
                event.setup['namespace_packages'].append('.'.join(package_bits[0:i]))

        package_dir = None  # useless, but helps showing the scope of this var
        for i in range(1, len(package_bits) + 1):
            _bits = package_bits[0:i]
            package_dir = os.path.join(*_bits)
            if not os.path.exists(package_dir):
                os.makedirs(package_dir)

            package_init_file = os.path.join(package_dir, '__init__.py')
            if not os.path.exists(package_init_file):
                is_namespace = i < len(package_bits)
                self.render_file(package_init_file, 'python/package_init.py.j2', {'is_namespace': is_namespace})

        # render version file, try to import version from version.txt if available.
        try:
            with open('version.txt') as version_file:
                version = version_file.read().strip()
        except IOError:
            version = '0.0.0'
        self.render_file_inline(os.path.join(package_dir, '_version.py'), "__version__ = '{}'".format(version))

        setup = event.config['python'].get_setup()

        context = {
            'url': setup.pop('url', 'http://example.com/'),
            'download_url': setup.pop('download_url', 'http://example.com/'),
        }

        for k, v in context.items():
            context[k] = context[k].format(
                name=setup['name'],
                user=getuser(),
                version='{version}',
            )

        context.update(
            {
                'data_files': setup.pop('data_files', {}),
                'entry_points': setup.pop('entry_points', {}),
                'extras_require': event.config['python'].get('extras_require'),
                'install_requires': event.config['python'].get('install_requires'),
                'setup': setup,
            }
        )

        # Render (with overwriting) the allmighty setup.py
        self.render_file('setup.py', 'python/setup.py.j2', context, override=True)

    @subscribe('medikit.on_end', priority=ABSOLUTE_PRIORITY)
    def on_end(self, event):
        pip_command = get_pip_command()
        pip_options, _ = pip_command.parse_args([])
        session = pip_command._build_session(pip_options)
        repository = PyPIRepository(pip_options, session)

        for extra in itertools.chain((None, ), event.config['python'].get_extras()):
            tmpfile = tempfile.NamedTemporaryFile(mode='wt', delete=False)
            if extra:
                tmpfile.write('\n'.join(event.config['python'].get_requirements(extra=extra)))
            else:
                tmpfile.write('\n'.join(event.config['python'].get_requirements()))
            tmpfile.flush()
            constraints = list(
                parse_requirements(
                    tmpfile.name, finder=repository.finder, session=repository.session, options=pip_options
                )
            )
            resolver = Resolver(constraints, repository, prereleases=False, clear_caches=False, allow_unsafe=False)

            self.render_file_inline(
                'requirements{}.txt'.format('-' + extra if extra else ''),
                '\n'.join(
                    (
                        '-e .{}'.format('[' + extra + ']' if extra else ''),
                        *sorted(
                            format_requirement(req) for req in resolver.resolve(max_rounds=10)
                            if req.name != event.config['python'].get('name')
                        ),
                    )
                )
            )


__feature__ = PythonFeature



"""
GNU Make support.

"""

import datetime
import itertools
import textwrap
from collections import deque

import medikit
from medikit.events import subscribe
from medikit.feature import Feature, HIGH_PRIORITY
from medikit.structs import Script
from whistle import Event


class Makefile(object):
    @property
    def targets(self):
        for key in self._target_order:
            yield key, self._target_values[key]

    @property
    def environ(self):
        return self._env_values

    def __init__(self):
        self._env_order, self._env_values, self._env_assignment_operators = deque(), {}, {}
        self._target_order, self._target_values = deque(), {}
        self.phony = set()

    def __delitem__(self, key):
        self._env_order.remove(key)
        del self._env_values[key]

    def __getitem__(self, item):
        return self._env_values[item]

    def __setitem__(self, key, value):
        self._env_values[key] = value
        if not key in self._env_order:
            self._env_order.append(key)

    def __iter__(self):
        for key in self._env_order:
            yield key, self._env_values[key]

    def __len__(self):
        return len(self._env_order)

    def __str__(self):
        content = [
            '# Generated by Medikit ' + medikit.__version__ + ' on ' + str(datetime.datetime.now().date()) + '.',
            '# All changes will be overriden.',
            '',
        ]

        if len(self):
            for k, v in self:
                v = textwrap.dedent(str(v)).strip()
                v = v.replace('\n', ' \\\n' + ' ' * (len(k) + 4))
                content.append('{} {} {}'.format(k, self._env_assignment_operators.get(k, '?='), v))
            content.append('')

        if len(self.phony):
            content.append('.PHONY: ' + ' '.join(sorted(self.phony)))
            content.append('')

        for target, details in self.targets:
            deps, rule, doc = details
            if doc:
                for line in doc.split('\n'):
                    content.append('# ' + line)
            content.append('{}: {}'.format(target, ' '.join(deps)).strip())

            script = textwrap.dedent(str(rule)).strip()

            for line in script.split('\n'):
                content.append('\t' + line)

            content.append('')

        return '\n'.join(content)

    def add_target(self, target, rule, deps=None, phony=False, first=False, doc=None):
        if target in self._target_order:
            raise RuntimeError('Duplicate definition for make target «{}».'.format(target))

        if isinstance(rule, str):
            rule = Script(rule)

        self._target_values[target] = (
            deps or list(),
            rule,
            textwrap.dedent(doc or '').strip(),
        )
        self._target_order.appendleft(target) if first else self._target_order.append(target)

        if phony:
            self.phony.add(target)

    def get_target(self, target):
        return self._target_values[target][1]

    def set_deps(self, target, deps=None):
        self._target_values[target] = (
            deps or list(),
            self._target_values[target][1],
            self._target_values[target][2],
        )

    def set_assignment_operator(self, key, value):
        assert value in ('?=', '=', '+=', ':=', '::=', '!='), 'Invalid operator'
        self._env_assignment_operators[key] = value

    def setleft(self, key, value):
        self._env_values[key] = value
        if not key in self._env_order:
            self._env_order.appendleft(key)

    def updateleft(self, *lst):
        for key, value in reversed(lst):
            self.setleft(key, value)


class MakefileEvent(Event):
    def __init__(self, package_name, makefile):
        self.package_name = package_name
        self.makefile = makefile
        super(MakefileEvent, self).__init__()


class InstallScript(Script):
    def __init__(self, script=None):
        super(InstallScript, self).__init__(script)

        self.before_install = []
        self.install = self.script
        self.after_install = []

    def __iter__(self):
        yield 'if [ -z "$(QUICK)" ]; then \\'
        for line in map(
            lambda x: '    {} ; \\'.format(x), itertools.chain(self.before_install, self.install, self.after_install)
        ):
            yield line
        yield 'fi'


class CleanScript(Script):
    remove = [
        'build',
        'dist',
        '*.egg-info',
    ]

    def __iter__(self):
        yield 'rm -rf {}'.format(' '.join(self.remove))


class MakeConfig(Feature.Config):
    def __init__(self):
        self.include_medikit_targets = True

    def disable_medikit_targets(self):
        self.include_medikit_targets = False


class MakeFeature(Feature):
    Config = MakeConfig

    def configure(self):
        self.makefile = Makefile()

    @subscribe('medikit.on_start', priority=HIGH_PRIORITY)
    def on_start(self, event):
        """
        :param ProjectEvent event:
        """
        for k in event.variables:
            self.makefile[k.upper()] = event.variables[k]

        self.makefile.updateleft((
            'QUICK',
            '',
        ), )

        self.makefile.add_target(
            'install', InstallScript(), phony=True, doc='''Installs the local project dependencies.'''
        )
        self.makefile.add_target(
            'install-dev',
            InstallScript(),
            phony=True,
            doc='''Installs the local project dependencies, including development-only libraries.'''
        )
        self.makefile.add_target('clean', CleanScript(), phony=True, doc='''Cleans up the local mess.''')

        if event.config['make'].include_medikit_targets:
            self.makefile.add_target(
                'update',
                '''
                python -c 'import medikit; print(medikit.__version__)' || pip install medikit;
                $(PYTHON) -m medikit update
            ''',
                phony=True,
                doc='''Update project artifacts using medikit, after installing it eventually.'''
            )

            self.makefile.add_target(
                'update-requirements',
                '''
                rm -rf requirements*.txt
                $(MAKE) update
            ''',
                phony=True,
                doc=
                '''Remove requirements files and update project artifacts using medikit, after installing it eventually.'''
            )

        self.dispatcher.dispatch(__name__ + '.on_generate', MakefileEvent(event.setup['name'], self.makefile))

        self.render_file_inline('Makefile', self.makefile.__str__(), override=True)


__feature__ = MakeFeature



# coding: utf-8

from __future__ import absolute_import, print_function, unicode_literals

DEFAULT_FEATURES = {
    'git',
    'make',
}

DEFAULT_FILES = {
    'requirements',
    'requirements-dev',
    'classifiers',
    'version',
}

YAPF_STYLE_CONFIG = 'pep8'



from stevedore import ExtensionManager

from medikit.settings import DEFAULT_FEATURES
from medikit.utils import format_file_content


class ConfigurationRegistry():
    def __init__(self):
        self._configs = {}
        self._features = {}

        def register_feature(ext):
            self._features[ext.name] = ext.plugin

        mgr = ExtensionManager(namespace='medikit.feature')
        mgr.map(register_feature)

    def __getitem__(self, item):
        return self._configs[item]

    def keys(self):
        return self._configs.keys()

    def require(self, *args):
        if len(args) == 1:
            return self._require(args[0])
        elif len(args) > 1:
            return tuple(map(self._require, args))
        raise ValueError('Empty.')

    def _require(self, name):
        if not name in self._features:
            raise ValueError('Unknown feature {!r}.'.format(name))

        if name not in self._configs:
            self._configs[name] = self._features[name].Config()

        return self._configs[name]


def read_configuration(dispatcher, filename, variables, features, files, setup):
    config = ConfigurationRegistry()

    import medikit
    _require_backup = medikit.require
    medikit.listen = dispatcher.listen
    medikit.require = config.require
    # todo runpy ?
    with open(filename) as f:
        code = compile(f.read(), filename, 'exec')
    ctx = {'listen': dispatcher.listen}
    exec(code, ctx)
    medikit.require = _require_backup

    for k in variables.keys():
        if k in ctx:
            variables[k] = ctx[k]

    for feature in DEFAULT_FEATURES:
        config.require(feature)

    # old, deprecated way ...
    for feature in set(ctx.pop('enable_features', ())) - set(ctx.pop('disable_features', ())):
        config.require(feature)

    # current way ...
    features = features | set(config.keys())

    for k in files.keys():
        if k in ctx:
            files[k] = format_file_content(ctx[k])

    for k in setup.keys():
        try:
            setup[k] = config['python'].get(k)
        except KeyError:
            if k in ctx:  # BC
                setup[k] = ctx[k]

        if setup[k] is None:
            raise ValueError('You must provide a value for the setup entry "{}" in your Projectfile.'.format(k))

    return variables, features, files, setup, config


_all_features = {}


def load_features():
    if not _all_features:

        def register_feature(ext, all_features=_all_features):
            all_features[ext.name] = ext.plugin

        mgr = ExtensionManager(namespace='medikit.feature')
        mgr.map(register_feature)

    return _all_features



"""
The «django» feature adds the django framework to your project.

"""

import os
import random

from medikit.events import subscribe
from . import Feature, SUPPORT_PRIORITY

random = random.SystemRandom()


def generate_secret_key():
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*(-_=+)'
    return ''.join(random.choice(chars) for i in range(64))


class DjangoConfig(Feature.Config):
    """ Configuration class for the “django” feature. """

    use_jinja2 = True
    """Whether or not to use the Jinja2 templating language (default: True)."""

    use_whitenoise = True
    """Whether or not to use Whitenoise for the static files (default: True)."""

    version = '==2.0a1'
    """Which django version requirement do you want? (default: ==2.0.a1)"""

    def __init__(self):
        self.use_jinja2 = self.use_jinja2
        self.use_whitenoise = self.use_whitenoise
        self.version = self.version


class DjangoFeature(Feature):
    requires = {'python'}

    Config = DjangoConfig

    @subscribe('medikit.feature.python.on_generate')
    def on_python_generate(self, event):
        event.config['python'].add_requirements('django ' + event.config['django'].version)

        if event.config['django'].use_jinja2:
            event.config['python'].add_requirements('Jinja2 >=2.9,<2.10', )

        if event.config['django'].use_whitenoise:
            event.config['python'].add_requirements(
                'brotlipy >=0.7,<0.8',
                'whitenoise >=3.3,<3.4',
            )

        event.config['python'].add_requirements('django-extensions >=1.9,<1.10', )

    @subscribe('medikit.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile
        makefile['DJANGO'] = '$(PYTHON) manage.py'
        makefile.add_target('runserver', '''$(DJANGO) runserver''', deps=('install-dev', ), phony=True)

    @subscribe('medikit.on_start')
    def on_start(self, event):

        context = {
            **event.setup,
            'config_package': 'config',
            'secret_key': generate_secret_key(),
            'use_jinja2': event.config['django'].use_jinja2,
            'use_whitenoise': event.config['django'].use_whitenoise,
        }

        name = context['name']

        # Create configuration
        config_path = 'config'
        if not os.path.exists(config_path):
            os.makedirs(config_path)

        self.render_file('manage.py', 'django/manage.py.j2', context, executable=True, force_python=True)
        self.render_file(os.path.join(config_path, 'settings.py'), 'django/settings.py.j2', context, force_python=True)
        self.render_file(os.path.join(config_path, 'urls.py'), 'django/urls.py.j2', context, force_python=True)
        self.render_file(os.path.join(config_path, 'wsgi.py'), 'django/wsgi.py.j2', context, force_python=True)

        self.dispatcher.dispatch('medikit.feature.django.on_configure')

        if context['use_jinja2']:
            templates_dir = os.path.join(name, 'jinja2', context['name'])
            if not os.path.exists(templates_dir):
                os.makedirs(templates_dir)

        static_dir = os.path.join(name, 'static')
        if not os.path.exists(static_dir):
            os.makedirs(static_dir)
        self.render_empty_files(os.path.join(static_dir, 'favicon.ico'))

    '''
    XXX todo use yapf to adjust settings (for example, add some apps).
    @subscribe('medikit.feature.django.on_configure')
    def on_django_configure(self, event):
        original_source, newline, encoding = yapf_api.ReadFile('config/settings.py')
        print(original_source, newline, encoding)
        reformatted_code, encoding, has_change = yapf_api.FormatFile(
            'config/settings.py',
            in_place=True,
        )
        print(reformatted_code, encoding, has_change)
    '''


__feature__ = DjangoFeature



"""
Not yet functional.
"""
from medikit.events import subscribe
from medikit.feature import Feature


class DockerFeature(Feature):
    @subscribe('medikit.feature.make.on_generate')
    def on_make_generate(self, event):
        event.makefile['DOCKER'] = "$(shell which docker)"
        event.makefile['DOCKER_PUSH'] = "$(DOCKER) push"
        event.makefile['DOCKER_REGISTRY'] = ""
        event.makefile['DOCKER_NAME'] = ""
        event.makefile['DOCKER_IMAGE'] = ""


__feature__ = DockerFeature



#!/usr/bin/env python
# coding: utf-8

from __future__ import absolute_import, print_function

import argparse
import logging
import os
import sys
from collections import OrderedDict
from contextlib import contextmanager

from blessings import Terminal

from medikit.config import read_configuration, load_features
from medikit.events import LoggingDispatcher, ProjectEvent
from medikit.feature import ProjectInitializer
from medikit.settings import DEFAULT_FEATURES, DEFAULT_FILES
from medikit.logging import getLogger

# Globals

logger = getLogger()

t = Terminal()


def _read_configuration(dispatcher, config_filename):
    """
    Prepare the python context and delegate to the real configuration reader (see config.py)

    :param EventDispatcher dispatcher:
    :return tuple: (variables, features, files, setup)
    """
    if not os.path.exists(config_filename):
        raise IOError('Could not find project description file (looked in {})'.format(config_filename))

    variables = OrderedDict()

    files = {filename: '' for filename in DEFAULT_FILES}

    setup = OrderedDict(
        (
            (
                'name',
                None,
            ),
            (
                'description',
                None,
            ),
            (
                'license',
                None,
            ),
            (
                'entry_points',
                {},
            ),
            (
                'install_requires',
                [],
            ),
            (
                'extras_require',
                {},
            ),
            (
                'data_files',
                [],
            ),
            (
                'url',
                'http://example.com/',
            ),
            (
                'download_url',
                'http://example.com/',
            ),
        )
    )

    features = set(DEFAULT_FEATURES)

    return read_configuration(dispatcher, config_filename, variables, features, files, setup)


def main(args=None):
    parser = argparse.ArgumentParser()

    parser.add_argument('--config', '-c', default='Projectfile')
    parser.add_argument('--verbose', '-v', action='store_true', default=False)

    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    # todo aliases for update/init
    # warning: http://bugs.python.org/issue9234

    parser_init = subparsers.add_parser('init', help='Initialize a new project.')
    parser_init.set_defaults(handler=handle_init)
    parser_init.add_argument('target')
    parser_init.add_argument('--name')
    parser_init.add_argument('--description')
    parser_init.add_argument('--license')
    parser_init.add_argument('--feature', '-f', action='append', dest='features')

    parser_update = subparsers.add_parser('update', help='Update current project.')
    parser_update.set_defaults(handler=handle_update)

    options = parser.parse_args(args if args is not None else sys.argv[1:])
    if options.verbose:
        logger.setLevel(logging.DEBUG)

    options = vars(options)
    handler = options.pop('handler')

    config_filename = os.path.join(os.getcwd(), options.pop('target', '.'), options.pop('config'))

    return handler(config_filename, **options)


@contextmanager
def _change_working_directory(path):
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


def handle_init(config_filename, **options):
    if os.path.exists(config_filename):
        raise IOError(
            'No config should be present in current directory to initialize (found {})'.format(config_filename)
        )

    config_dirname = os.path.dirname(config_filename)
    if not os.path.exists(config_dirname):
        os.makedirs(config_dirname)

    # Fast and dirty implementation
    # TODO
    # - input validation
    # - getting input from env/git conf (author...),
    # - dispatching something in selected features so maybe they can suggest deps
    # - deps selection
    # - ...
    with _change_working_directory(config_dirname):
        dispatcher = LoggingDispatcher()
        initializer = ProjectInitializer(dispatcher, options)
        initializer.execute()
        return handle_update(config_filename)


def handle_update(config_filename, **kwargs):
    dispatcher = LoggingDispatcher()

    variables, features, files, setup, config = _read_configuration(dispatcher, config_filename)

    feature_instances = {}
    logger.info(
        'Updating {} with {} features'.format(
            t.bold(setup['name']), ', '.join(t.bold(t.green(feature_name)) for feature_name in sorted(features))
        )
    )

    all_features = load_features()

    sorted_features = sorted(features)  # sort to have a predictable display order
    for feature_name in sorted_features:
        logger.debug('Initializing feature {}...'.format(t.bold(t.green(feature_name))))
        try:
            feature = all_features[feature_name]
        except KeyError as exc:
            logger.exception('Feature "{}" not found.'.format(feature_name))

        if feature:
            feature_instances[feature_name] = feature(dispatcher)

            for req in feature_instances[feature_name].requires:
                if not req in sorted_features:
                    raise RuntimeError('Unmet dependency: {} requires {}.'.format(feature_name, req))

            for con in feature_instances[feature_name].conflicts:
                if con in sorted_features:
                    raise RuntimeError('Conflicting dependency: {} conflicts with {}.'.format(con, feature_name))
        else:
            raise RuntimeError('Required feature {} not found.'.format(feature_name))

    event = ProjectEvent(config=config)
    event.variables, event.files, event.setup = variables, files, setup

    # todo: add listener dump list in debug/verbose mode ?

    event = dispatcher.dispatch('medikit.on_start', event)

    dispatcher.dispatch('medikit.on_end', event)

    logger.info(u'Done.')


if __name__ == '__main__':
    main()
